#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xmd.h>
#include <X11/SM/SMlib.h>
#include <X11/Xatom.h>

#define	_NET_WM_STATE_REMOVE	0
#define	_NET_WM_STATE_ADD	1
#define	_NET_WM_STATE_TOGGLE	2

typedef struct _SYSMON
{
    GtkWidget *window;
    GtkWidget *eventbox;
    GtkWidget *label;

    GtkStyle *style;

    GdkColor bg_color;

    gulong user, nice, sys, idle, w, x, y;
    gulong prev_user, prev_nice, prev_sys, prev_idle, prev_w, prev_x, prev_y;
} SYSMON;

SYSMON gui;

/* obsolete */
gulong kb_main_shared;
/* old but still kicking -- the important stuff */
gulong kb_main_buffers;
gulong kb_main_cached;
gulong kb_main_free;
gulong kb_main_total;
gulong kb_swap_free;
gulong kb_swap_total;
/* recently introduced */
gulong kb_high_free;
gulong kb_high_total;
gulong kb_low_free;
gulong kb_low_total;
/* 2.4.xx era */
gulong kb_active;
gulong kb_inact_laundry;
gulong kb_inact_dirty;
gulong kb_inact_clean;
gulong kb_inact_target;
gulong kb_swap_cached;  /* late 2.4 only */
/* derived values */
gulong kb_swap_used;
gulong kb_main_used;
/* 2.5.41+ */
gulong kb_writeback;
gulong kb_dirty;
gulong kb_inactive;

void meminfo ()
{
    char buf[1024];
    FILE *fd;

    if ((fd = fopen("/proc/meminfo", "r")) == NULL)
        return;

    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_main_total);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_main_free);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_main_buffers);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_main_cached);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_swap_cached);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_active);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_inactive);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_high_total);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_high_free);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_low_total);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_low_free);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_swap_total);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_swap_free);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_dirty);
    if ((fgets(buf, sizeof(buf), fd)) != NULL)
        sscanf(buf, "%*s %lu", &kb_writeback);

    fclose (fd);

    kb_inactive = ~0UL;

    if(!kb_low_total){  /* low==main except with large-memory support */
        kb_low_total = kb_main_total;
        kb_low_free  = kb_main_free;
    }
    if(kb_inactive==~0UL){
        kb_inactive = kb_inact_dirty + kb_inact_clean + kb_inact_laundry;
    }
    kb_swap_used = kb_swap_total - kb_swap_free;
    kb_main_used = kb_main_total - kb_main_free;
}

void destroy (GtkWidget *widget,
              gpointer   data )
{
    gtk_main_quit ();
}

// returns percent usage
gfloat get_cpu_usage (gboolean calculate)
{
    FILE *f;
    gchar *s, buf[1024];
    gint n;

    if ((f = fopen("/proc/stat", "r")) == NULL)
        return;

    if (calculate == TRUE)
    {
        if ((fgets(buf, sizeof(buf), f)) != NULL)
        {
            n = sscanf(buf, "%*s %lu %lu %lu %lu %lu %lu %lu %lu",
                            &gui.user, &gui.nice, &gui.sys, &gui.idle, &gui.w, &gui.x, &gui.y);
        }

        glong u_frme = gui.user - gui.prev_user;
        glong s_frme = gui.sys - gui.prev_sys;
        glong n_frme = gui.nice - gui.prev_nice;
        glong i_frme = gui.idle - gui.prev_idle;
        glong w_frme = gui.w - gui.prev_w;
        glong x_frme = gui.x - gui.prev_x;
        glong y_frme = gui.y - gui.prev_y;

        if (i_frme < 0) i_frme = 0;

        glong tot_frme = u_frme + s_frme + n_frme + i_frme + w_frme + x_frme + y_frme;
        if (1 > tot_frme) tot_frme = 1;
        gfloat scale = 100.0 / (float)tot_frme;

        gui.prev_user = gui.user;
        gui.prev_sys = gui.sys;
        gui.prev_nice = gui.nice;
        gui.prev_idle = gui.idle;
        gui.prev_w = gui.w;
        gui.prev_x = gui.x;
        gui.prev_y = gui.y;

        fclose(f);

        return (gfloat)(u_frme + s_frme + n_frme + w_frme + x_frme + y_frme) * scale;
    }
    else
    {
        if ((fgets(buf, sizeof(buf), f)) != NULL)
        {
            n = sscanf(buf, "%*s %lu %lu %lu %lu %lu %lu %lu %lu",
                            &gui.prev_user, &gui.prev_nice, &gui.prev_sys, &gui.prev_idle, &gui.prev_w, &gui.prev_x, &gui.prev_y);
        }

        fclose(f);

        return 0;
    }
}

time_t get_uptime ()
{
    FILE *f;
    gulong l = 0;

    if ((f = fopen("/proc/uptime", "r")) != NULL)
    {
        fscanf(f, "%lu", &l);
        fclose(f);
    }
    return (time_t) l;
}

void update_stats ()
{
    const gint MAX_INFO_LENGTH = 1000;
    gchar info[MAX_INFO_LENGTH];

    /* Conversion constants. */
    const long minute = 60;
    const long hour = minute * 60;
    const long day = hour * 24;
    const double megabyte = 1024 * 1024;

    // uptime information
    time_t uptime = get_uptime();
    time_t updays = uptime / 86400;
    time_t uphours = (uptime - (updays * 86400)) /  3600;
    time_t upmins = (uptime - (updays * 86400) - (uphours * 3600)) / 60;

    // cpu usage information
    gfloat cpu_usage = get_cpu_usage(TRUE);
 
    // mem information
    meminfo();
    gulong buffers_plus_cached = kb_main_buffers + kb_main_cached;
    gint mem_free = (kb_main_free + buffers_plus_cached) >> 10;
    gint mem_total = kb_main_total >> 10;

    // swap information
    gint swap_total = kb_swap_total >> 10;
    gint swap_free = kb_swap_free >> 10;

    sprintf(info, "UPTIME: %ld DAYS, %ld:%02ld\nFREE SWAP: %d/%d\nFREE RAM: %d/%d\nCPU: %2.1f%\n",
        updays, uphours, upmins,
        swap_free, swap_total,
        mem_free, mem_total,
        cpu_usage
    );

    gtk_label_set_text (GTK_LABEL(gui.label), info);
}


void net_wm_state (gchar    *hint,
                   gboolean  state)
{
    XEvent xev;

    xev.type = ClientMessage;
    xev.xclient.type = ClientMessage;
    xev.xclient.window = GDK_WINDOW_XWINDOW(gui.window->window);
    xev.xclient.message_type = gdk_x11_get_xatom_by_name("_NET_WM_STATE");
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = state ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
    xev.xclient.data.l[1] = gdk_x11_get_xatom_by_name(hint);
    xev.xclient.data.l[2] = 0;

    XSendEvent(GDK_DISPLAY(), GDK_ROOT_WINDOW(), False, SubstructureNotifyMask, &xev);
}

void winop_options ()
{
    Display *display;
    Window window;
    Atom atoms[2];
    gint n = 0;

    display = GDK_WINDOW_XDISPLAY(gui.window->window);
    window  = GDK_WINDOW_XWINDOW(gui.window->window);

    atoms[0] = XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR", False);
    atoms[1] = XInternAtom(display, "_NET_WM_STATE_SKIP_PAGER", False);
    XChangeProperty(display, window, XInternAtom(display, "_NET_WM_STATE", False),
                    XA_ATOM, 32, PropModeReplace, (guchar *) atoms, 2);

    XWMHints mywmhints; 
    mywmhints.initial_state = WithdrawnState; 
    mywmhints.flags=StateHint;

    XSetWMHints(display, window, &mywmhints);
}

gboolean cb_map_event (GtkWidget *widget,
                       GdkEvent  *event,
                       gpointer   data)
{
    net_wm_state("_NET_WM_STATE_SKIP_PAGER", TRUE);
    net_wm_state("_NET_WM_STATE_SKIP_TASKBAR", TRUE);

    return FALSE;    gtk_widget_destroy(gui.window);
}

gint main (int   argc,
           char *argv[] )
{
    gtk_init (&argc, &argv);
    
    // Create a new window
    gui.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    
    // Disable window resizing
    gtk_window_set_resizable ((GtkWindow *) gui.window, FALSE);

    // Disable window decoration
    gtk_window_set_decorated((GtkWindow *) gui.window, FALSE);
    
    /* Here we connect the "destroy" event to a signal handler.  
     * This event occurs when we call gtk_widget_destroy() on the window,
     * or if we return FALSE in the "delete_event" callback. */
    g_signal_connect (G_OBJECT (gui.window), "destroy",
		      G_CALLBACK (destroy), NULL);

    // Creates a new event box
    gui.eventbox = gtk_event_box_new();

    // Creates a new label with text default to sysmon
    gui.label = gtk_label_new("sysmonitor - tdk");

    // Set the widget name so the label's font can be customized through .gtkrc-2.0
    gtk_widget_set_name (gui.label, "system_stats");

    // Make background white to match my background
    gdk_color_parse("white", &gui.bg_color);
    gtk_widget_modify_bg (gui.eventbox, GTK_STATE_NORMAL, &gui.bg_color);

    // Make label right aligned
    gtk_label_set_justify ((GtkLabel *)gui.label, GTK_JUSTIFY_RIGHT);

    g_signal_connect(G_OBJECT(gui.window), "map_event",
			G_CALLBACK(cb_map_event), NULL);

    // TODO: remove once finished coding
    // TODO: need alternative method for closing the stats... maybe right click menu or something
/*
    g_signal_connect_swapped (G_OBJECT (gui.eventbox), "button_press_event",
			      G_CALLBACK (gtk_widget_destroy),
                              G_OBJECT (gui.window));
*/  
    // This packs the label into the event box
    gtk_container_add (GTK_CONTAINER (gui.eventbox), gui.label);

    // This packs the event box into the window (a gtk container).
    gtk_container_add (GTK_CONTAINER (gui.window), gui.eventbox);
    
    // Move the window to where I want it to be
    gtk_window_move ((GtkWindow *) gui.window, 1200, 40);
    gtk_window_set_default_size ((GtkWindow *) gui.window, 300, 300);

    // fill in values for the old stat
    get_cpu_usage(FALSE);

    gint interval = 3 * 1000; // update every 3 seconds
    interval = interval * 60 / 63;	// Compensate for overhead XXX
    gint timeoutid = gtk_timeout_add(interval, (GtkFunction) update_stats, NULL);

    // Display the widgets
    gtk_widget_show (gui.label);
    gtk_widget_show (gui.eventbox);
    gtk_widget_show (gui.window);

    // Make system stats appear on all workspaces
    gtk_window_stick(GTK_WINDOW(gui.window));

    winop_options();

    // Wait for mouse events...
    gtk_main ();

    gtk_timeout_remove(timeoutid);
    
    return 0;
}
